package fp

import (
	"math/bits"
)

// Mul z = x * y (mod q)
//
// x and y must be strictly inferior to q
func (z *Element) Mul(x, y *Element) *Element {

	// Implements CIOS multiplication -- section 2.3.2 of Tolga Acar's thesis
	// https://www.microsoft.com/en-us/research/wp-content/uploads/1998/06/97Acar.pdf
	//
	// The algorithm:
	//
	// for i=0 to N-1
	// 		C := 0
	// 		for j=0 to N-1
	// 			(C,t[j]) := t[j] + x[j]*y[i] + C
	// 		(t[N+1],t[N]) := t[N] + C
	//
	// 		C := 0
	// 		m := t[0]*q'[0] mod D
	// 		(C,_) := t[0] + m*q[0]
	// 		for j=1 to N-1
	// 			(C,t[j-1]) := t[j] + m*q[j] + C
	//
	// 		(C,t[N-1]) := t[N] + C
	// 		t[N] := t[N+1] + C
	//
	// → N is the number of machine words needed to store the modulus q
	// → D is the word size. For example, on a 64-bit architecture D is 2	64
	// → x[i], y[i], q[i] is the ith word of the numbers x,y,q
	// → q'[0] is the lowest word of the number -q⁻¹ mod r. This quantity is pre-computed, as it does not depend on the inputs.
	// → t is a temporary array of size N+2
	// → C, S are machine words. A pair (C,S) refers to (hi-bits, lo-bits) of a two-word number
	//
	// As described in the paper appendix, we can get rid of one carry chain and simplify:
	//
	// for i=0 to N-1
	// 		(A,t[0]) := t[0] + x[0]*y[i]
	// 		m := t[0]*q'[0] mod W
	// 		C,_ := t[0] + m*q[0]
	// 		for j=1 to N-1
	// 			(A,t[j])  := t[j] + x[j]*y[i] + A
	// 			(C,t[j-1]) := t[j] + m*q[j] + C
	//
	// 		t[N-1] = C + A
	//
	// This optimization saves 5N + 2 additions in the algorithm, and can be used whenever the highest bit
	// of the modulus is zero (and not all of the remaining bits are set).

	var t0, t1, t2, t3, t4, t5 uint64
	var u0, u1, u2, u3, u4, u5 uint64
	var v0, v1, v2, v3, v4, v5 uint64
	v0 = y[0]
	v1 = y[1]
	v2 = y[2]
	v3 = y[3]
	v4 = y[4]
	v5 = y[5]
	{
		var c0, c1, c2 uint64
		v := x[0]
		u0, t0 = bits.Mul64(v, v0)
		u1, t1 = bits.Mul64(v, v1)
		u2, t2 = bits.Mul64(v, v2)
		u3, t3 = bits.Mul64(v, v3)
		u4, t4 = bits.Mul64(v, v4)
		u5, t5 = bits.Mul64(v, v5)
		t1, c0 = bits.Add64(u0, t1, 0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, 0, c0)

		m := qInvNeg * t0

		u0, c1 = bits.Mul64(m, q0)
		_, c0 = bits.Add64(t0, c1, 0)
		u1, c1 = bits.Mul64(m, q1)
		t0, c0 = bits.Add64(t1, c1, c0)
		u2, c1 = bits.Mul64(m, q2)
		t1, c0 = bits.Add64(t2, c1, c0)
		u3, c1 = bits.Mul64(m, q3)
		t2, c0 = bits.Add64(t3, c1, c0)
		u4, c1 = bits.Mul64(m, q4)
		t3, c0 = bits.Add64(t4, c1, c0)
		u5, c1 = bits.Mul64(m, q5)

		t4, c0 = bits.Add64(0, c1, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)

	}
	{
		var c0, c1, c2 uint64
		v := x[1]
		u0, c1 = bits.Mul64(v, v0)
		t0, c0 = bits.Add64(c1, t0, 0)
		u1, c1 = bits.Mul64(v, v1)
		t1, c0 = bits.Add64(c1, t1, c0)
		u2, c1 = bits.Mul64(v, v2)
		t2, c0 = bits.Add64(c1, t2, c0)
		u3, c1 = bits.Mul64(v, v3)
		t3, c0 = bits.Add64(c1, t3, c0)
		u4, c1 = bits.Mul64(v, v4)
		t4, c0 = bits.Add64(c1, t4, c0)
		u5, c1 = bits.Mul64(v, v5)
		t5, c0 = bits.Add64(c1, t5, c0)

		c2, _ = bits.Add64(0, 0, c0)
		t1, c0 = bits.Add64(u0, t1, 0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		m := qInvNeg * t0

		u0, c1 = bits.Mul64(m, q0)
		_, c0 = bits.Add64(t0, c1, 0)
		u1, c1 = bits.Mul64(m, q1)
		t0, c0 = bits.Add64(t1, c1, c0)
		u2, c1 = bits.Mul64(m, q2)
		t1, c0 = bits.Add64(t2, c1, c0)
		u3, c1 = bits.Mul64(m, q3)
		t2, c0 = bits.Add64(t3, c1, c0)
		u4, c1 = bits.Mul64(m, q4)
		t3, c0 = bits.Add64(t4, c1, c0)
		u5, c1 = bits.Mul64(m, q5)

		t4, c0 = bits.Add64(0, c1, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)

	}
	{
		var c0, c1, c2 uint64
		v := x[2]
		u0, c1 = bits.Mul64(v, v0)
		t0, c0 = bits.Add64(c1, t0, 0)
		u1, c1 = bits.Mul64(v, v1)
		t1, c0 = bits.Add64(c1, t1, c0)
		u2, c1 = bits.Mul64(v, v2)
		t2, c0 = bits.Add64(c1, t2, c0)
		u3, c1 = bits.Mul64(v, v3)
		t3, c0 = bits.Add64(c1, t3, c0)
		u4, c1 = bits.Mul64(v, v4)
		t4, c0 = bits.Add64(c1, t4, c0)
		u5, c1 = bits.Mul64(v, v5)
		t5, c0 = bits.Add64(c1, t5, c0)

		c2, _ = bits.Add64(0, 0, c0)
		t1, c0 = bits.Add64(u0, t1, 0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		m := qInvNeg * t0

		u0, c1 = bits.Mul64(m, q0)
		_, c0 = bits.Add64(t0, c1, 0)
		u1, c1 = bits.Mul64(m, q1)
		t0, c0 = bits.Add64(t1, c1, c0)
		u2, c1 = bits.Mul64(m, q2)
		t1, c0 = bits.Add64(t2, c1, c0)
		u3, c1 = bits.Mul64(m, q3)
		t2, c0 = bits.Add64(t3, c1, c0)
		u4, c1 = bits.Mul64(m, q4)
		t3, c0 = bits.Add64(t4, c1, c0)
		u5, c1 = bits.Mul64(m, q5)

		t4, c0 = bits.Add64(0, c1, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)

	}
	{
		var c0, c1, c2 uint64
		v := x[3]
		u0, c1 = bits.Mul64(v, v0)
		t0, c0 = bits.Add64(c1, t0, 0)
		u1, c1 = bits.Mul64(v, v1)
		t1, c0 = bits.Add64(c1, t1, c0)
		u2, c1 = bits.Mul64(v, v2)
		t2, c0 = bits.Add64(c1, t2, c0)
		u3, c1 = bits.Mul64(v, v3)
		t3, c0 = bits.Add64(c1, t3, c0)
		u4, c1 = bits.Mul64(v, v4)
		t4, c0 = bits.Add64(c1, t4, c0)
		u5, c1 = bits.Mul64(v, v5)
		t5, c0 = bits.Add64(c1, t5, c0)

		c2, _ = bits.Add64(0, 0, c0)
		t1, c0 = bits.Add64(u0, t1, 0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		m := qInvNeg * t0

		u0, c1 = bits.Mul64(m, q0)
		_, c0 = bits.Add64(t0, c1, 0)
		u1, c1 = bits.Mul64(m, q1)
		t0, c0 = bits.Add64(t1, c1, c0)
		u2, c1 = bits.Mul64(m, q2)
		t1, c0 = bits.Add64(t2, c1, c0)
		u3, c1 = bits.Mul64(m, q3)
		t2, c0 = bits.Add64(t3, c1, c0)
		u4, c1 = bits.Mul64(m, q4)
		t3, c0 = bits.Add64(t4, c1, c0)
		u5, c1 = bits.Mul64(m, q5)

		t4, c0 = bits.Add64(0, c1, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)

	}
	{
		var c0, c1, c2 uint64
		v := x[4]
		u0, c1 = bits.Mul64(v, v0)
		t0, c0 = bits.Add64(c1, t0, 0)
		u1, c1 = bits.Mul64(v, v1)
		t1, c0 = bits.Add64(c1, t1, c0)
		u2, c1 = bits.Mul64(v, v2)
		t2, c0 = bits.Add64(c1, t2, c0)
		u3, c1 = bits.Mul64(v, v3)
		t3, c0 = bits.Add64(c1, t3, c0)
		u4, c1 = bits.Mul64(v, v4)
		t4, c0 = bits.Add64(c1, t4, c0)
		u5, c1 = bits.Mul64(v, v5)
		t5, c0 = bits.Add64(c1, t5, c0)

		c2, _ = bits.Add64(0, 0, c0)
		t1, c0 = bits.Add64(u0, t1, 0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		m := qInvNeg * t0

		u0, c1 = bits.Mul64(m, q0)
		_, c0 = bits.Add64(t0, c1, 0)
		u1, c1 = bits.Mul64(m, q1)
		t0, c0 = bits.Add64(t1, c1, c0)
		u2, c1 = bits.Mul64(m, q2)
		t1, c0 = bits.Add64(t2, c1, c0)
		u3, c1 = bits.Mul64(m, q3)
		t2, c0 = bits.Add64(t3, c1, c0)
		u4, c1 = bits.Mul64(m, q4)
		t3, c0 = bits.Add64(t4, c1, c0)
		u5, c1 = bits.Mul64(m, q5)

		t4, c0 = bits.Add64(0, c1, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)

	}
	{
		var c0, c1, c2 uint64
		v := x[5]
		u0, c1 = bits.Mul64(v, v0)
		t0, c0 = bits.Add64(c1, t0, 0)
		u1, c1 = bits.Mul64(v, v1)
		t1, c0 = bits.Add64(c1, t1, c0)
		u2, c1 = bits.Mul64(v, v2)
		t2, c0 = bits.Add64(c1, t2, c0)
		u3, c1 = bits.Mul64(v, v3)
		t3, c0 = bits.Add64(c1, t3, c0)
		u4, c1 = bits.Mul64(v, v4)
		t4, c0 = bits.Add64(c1, t4, c0)
		u5, c1 = bits.Mul64(v, v5)
		t5, c0 = bits.Add64(c1, t5, c0)

		c2, _ = bits.Add64(0, 0, c0)
		t1, c0 = bits.Add64(u0, t1, 0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		m := qInvNeg * t0

		u0, c1 = bits.Mul64(m, q0)
		_, c0 = bits.Add64(t0, c1, 0)
		u1, c1 = bits.Mul64(m, q1)
		t0, c0 = bits.Add64(t1, c1, c0)
		u2, c1 = bits.Mul64(m, q2)
		t1, c0 = bits.Add64(t2, c1, c0)
		u3, c1 = bits.Mul64(m, q3)
		t2, c0 = bits.Add64(t3, c1, c0)
		u4, c1 = bits.Mul64(m, q4)
		t3, c0 = bits.Add64(t4, c1, c0)
		u5, c1 = bits.Mul64(m, q5)

		t4, c0 = bits.Add64(0, c1, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)

	}
	z[0] = t0
	z[1] = t1
	z[2] = t2
	z[3] = t3
	z[4] = t4
	z[5] = t5

	// if z ⩾ q → z -= q
	if !z.smallerThanModulus() {
		var b uint64
		z[0], b = bits.Sub64(z[0], q0, 0)
		z[1], b = bits.Sub64(z[1], q1, b)
		z[2], b = bits.Sub64(z[2], q2, b)
		z[3], b = bits.Sub64(z[3], q3, b)
		z[4], b = bits.Sub64(z[4], q4, b)
		z[5], _ = bits.Sub64(z[5], q5, b)
	}
	return z
}

// Square z = x * x (mod q)
//
// x must be strictly inferior to q
func (z *Element) Square(x *Element) *Element {
	// see Mul for algorithm documentation

	var t0, t1, t2, t3, t4, t5 uint64
	var u0, u1, u2, u3, u4, u5 uint64
	var lo0, lo1, lo2, lo3, lo4, lo5 uint64

	// note that if hi, _ = bits.Mul64() didn't generate
	// UMULH and MUL, (but just UMULH) we could use same pattern
	// as in mulRaw and reduce the stack space of this function (no need for lo..)
	{

		var c0, c2 uint64

		// for j=i+1 to N-1
		//     p,C,t[j] = 2*a[j]*a[i] + t[j] + (p,C)
		// A = C
		u0, lo1 = bits.Mul64(x[0], x[0])
		u1, t1 = bits.Mul64(x[1], x[0])
		u2, t2 = bits.Mul64(x[2], x[0])
		u3, t3 = bits.Mul64(x[3], x[0])
		u4, t4 = bits.Mul64(x[4], x[0])
		u5, t5 = bits.Mul64(x[5], x[0])

		// propagate lo, from t[j] to end, twice.
		t1, c0 = bits.Add64(t1, t1, 0)
		t2, c0 = bits.Add64(t2, t2, c0)
		t3, c0 = bits.Add64(t3, t3, c0)
		t4, c0 = bits.Add64(t4, t4, c0)
		t5, c0 = bits.Add64(t5, t5, c0)
		c2, _ = bits.Add64(c2, 0, c0)

		t0, c0 = bits.Add64(lo1, t0, 0)

		// propagate u0 + hi
		t1, c0 = bits.Add64(u0, t1, c0)
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		// hi again
		t2, c0 = bits.Add64(u1, t2, 0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(c2, u5, c0)

		// this part is unchanged.
		m := qInvNeg * t0
		u0, lo0 = bits.Mul64(m, q0)
		u1, lo1 = bits.Mul64(m, q1)
		u2, lo2 = bits.Mul64(m, q2)
		u3, lo3 = bits.Mul64(m, q3)
		u4, lo4 = bits.Mul64(m, q4)
		u5, lo5 = bits.Mul64(m, q5)
		_, c0 = bits.Add64(t0, lo0, 0)
		t0, c0 = bits.Add64(t1, lo1, c0)
		t1, c0 = bits.Add64(t2, lo2, c0)
		t2, c0 = bits.Add64(t3, lo3, c0)
		t3, c0 = bits.Add64(t4, lo4, c0)
		t4, c0 = bits.Add64(0, lo5, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)
	}
	{

		var c0, c2 uint64

		// for j=i+1 to N-1
		//     p,C,t[j] = 2*a[j]*a[i] + t[j] + (p,C)
		// A = C
		u1, lo1 = bits.Mul64(x[1], x[1])
		u2, lo2 = bits.Mul64(x[2], x[1])
		u3, lo3 = bits.Mul64(x[3], x[1])
		u4, lo4 = bits.Mul64(x[4], x[1])
		u5, lo5 = bits.Mul64(x[5], x[1])

		// propagate lo, from t[j] to end, twice.
		lo2, c0 = bits.Add64(lo2, lo2, 0)
		lo3, c0 = bits.Add64(lo3, lo3, c0)
		lo4, c0 = bits.Add64(lo4, lo4, c0)
		lo5, c0 = bits.Add64(lo5, lo5, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t2, c0 = bits.Add64(lo2, t2, 0)
		t3, c0 = bits.Add64(lo3, t3, c0)
		t4, c0 = bits.Add64(lo4, t4, c0)
		t5, c0 = bits.Add64(lo5, t5, c0)
		c2, _ = bits.Add64(c2, 0, c0)

		t1, c0 = bits.Add64(lo1, t1, 0)

		// propagate u1 + hi
		t2, c0 = bits.Add64(u1, t2, c0)
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		// hi again
		t3, c0 = bits.Add64(u2, t3, 0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(c2, u5, c0)

		// this part is unchanged.
		m := qInvNeg * t0
		u0, lo0 = bits.Mul64(m, q0)
		u1, lo1 = bits.Mul64(m, q1)
		u2, lo2 = bits.Mul64(m, q2)
		u3, lo3 = bits.Mul64(m, q3)
		u4, lo4 = bits.Mul64(m, q4)
		u5, lo5 = bits.Mul64(m, q5)
		_, c0 = bits.Add64(t0, lo0, 0)
		t0, c0 = bits.Add64(t1, lo1, c0)
		t1, c0 = bits.Add64(t2, lo2, c0)
		t2, c0 = bits.Add64(t3, lo3, c0)
		t3, c0 = bits.Add64(t4, lo4, c0)
		t4, c0 = bits.Add64(0, lo5, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)
	}
	{

		var c0, c2 uint64

		// for j=i+1 to N-1
		//     p,C,t[j] = 2*a[j]*a[i] + t[j] + (p,C)
		// A = C
		u2, lo2 = bits.Mul64(x[2], x[2])
		u3, lo3 = bits.Mul64(x[3], x[2])
		u4, lo4 = bits.Mul64(x[4], x[2])
		u5, lo5 = bits.Mul64(x[5], x[2])

		// propagate lo, from t[j] to end, twice.
		lo3, c0 = bits.Add64(lo3, lo3, 0)
		lo4, c0 = bits.Add64(lo4, lo4, c0)
		lo5, c0 = bits.Add64(lo5, lo5, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t3, c0 = bits.Add64(lo3, t3, 0)
		t4, c0 = bits.Add64(lo4, t4, c0)
		t5, c0 = bits.Add64(lo5, t5, c0)
		c2, _ = bits.Add64(c2, 0, c0)

		t2, c0 = bits.Add64(lo2, t2, 0)

		// propagate u2 + hi
		t3, c0 = bits.Add64(u2, t3, c0)
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		// hi again
		t4, c0 = bits.Add64(u3, t4, 0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(c2, u5, c0)

		// this part is unchanged.
		m := qInvNeg * t0
		u0, lo0 = bits.Mul64(m, q0)
		u1, lo1 = bits.Mul64(m, q1)
		u2, lo2 = bits.Mul64(m, q2)
		u3, lo3 = bits.Mul64(m, q3)
		u4, lo4 = bits.Mul64(m, q4)
		u5, lo5 = bits.Mul64(m, q5)
		_, c0 = bits.Add64(t0, lo0, 0)
		t0, c0 = bits.Add64(t1, lo1, c0)
		t1, c0 = bits.Add64(t2, lo2, c0)
		t2, c0 = bits.Add64(t3, lo3, c0)
		t3, c0 = bits.Add64(t4, lo4, c0)
		t4, c0 = bits.Add64(0, lo5, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)
	}
	{

		var c0, c2 uint64

		// for j=i+1 to N-1
		//     p,C,t[j] = 2*a[j]*a[i] + t[j] + (p,C)
		// A = C
		u3, lo3 = bits.Mul64(x[3], x[3])
		u4, lo4 = bits.Mul64(x[4], x[3])
		u5, lo5 = bits.Mul64(x[5], x[3])

		// propagate lo, from t[j] to end, twice.
		lo4, c0 = bits.Add64(lo4, lo4, 0)
		lo5, c0 = bits.Add64(lo5, lo5, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(lo4, t4, 0)
		t5, c0 = bits.Add64(lo5, t5, c0)
		c2, _ = bits.Add64(c2, 0, c0)

		t3, c0 = bits.Add64(lo3, t3, 0)

		// propagate u3 + hi
		t4, c0 = bits.Add64(u3, t4, c0)
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		// hi again
		t5, c0 = bits.Add64(u4, t5, 0)
		c2, _ = bits.Add64(c2, u5, c0)

		// this part is unchanged.
		m := qInvNeg * t0
		u0, lo0 = bits.Mul64(m, q0)
		u1, lo1 = bits.Mul64(m, q1)
		u2, lo2 = bits.Mul64(m, q2)
		u3, lo3 = bits.Mul64(m, q3)
		u4, lo4 = bits.Mul64(m, q4)
		u5, lo5 = bits.Mul64(m, q5)
		_, c0 = bits.Add64(t0, lo0, 0)
		t0, c0 = bits.Add64(t1, lo1, c0)
		t1, c0 = bits.Add64(t2, lo2, c0)
		t2, c0 = bits.Add64(t3, lo3, c0)
		t3, c0 = bits.Add64(t4, lo4, c0)
		t4, c0 = bits.Add64(0, lo5, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)
	}
	{

		var c0, c2 uint64

		// for j=i+1 to N-1
		//     p,C,t[j] = 2*a[j]*a[i] + t[j] + (p,C)
		// A = C
		u4, lo4 = bits.Mul64(x[4], x[4])
		u5, lo5 = bits.Mul64(x[5], x[4])

		// propagate lo, from t[j] to end, twice.
		lo5, c0 = bits.Add64(lo5, lo5, 0)
		c2, _ = bits.Add64(c2, 0, c0)
		t5, c0 = bits.Add64(lo5, t5, 0)
		c2, _ = bits.Add64(c2, 0, c0)

		t4, c0 = bits.Add64(lo4, t4, 0)

		// propagate u4 + hi
		t5, c0 = bits.Add64(u4, t5, c0)
		c2, _ = bits.Add64(u5, c2, c0)

		// hi again
		c2, _ = bits.Add64(c2, u5, 0)

		// this part is unchanged.
		m := qInvNeg * t0
		u0, lo0 = bits.Mul64(m, q0)
		u1, lo1 = bits.Mul64(m, q1)
		u2, lo2 = bits.Mul64(m, q2)
		u3, lo3 = bits.Mul64(m, q3)
		u4, lo4 = bits.Mul64(m, q4)
		u5, lo5 = bits.Mul64(m, q5)
		_, c0 = bits.Add64(t0, lo0, 0)
		t0, c0 = bits.Add64(t1, lo1, c0)
		t1, c0 = bits.Add64(t2, lo2, c0)
		t2, c0 = bits.Add64(t3, lo3, c0)
		t3, c0 = bits.Add64(t4, lo4, c0)
		t4, c0 = bits.Add64(0, lo5, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)
	}
	{

		var c0, c2 uint64

		// for j=i+1 to N-1
		//     p,C,t[j] = 2*a[j]*a[i] + t[j] + (p,C)
		// A = C
		u5, lo5 = bits.Mul64(x[5], x[5])

		// propagate lo, from t[j] to end, twice.

		t5, c0 = bits.Add64(lo5, t5, 0)

		// propagate u5 + hi
		c2, _ = bits.Add64(u5, c2, c0)

		// hi again

		// this part is unchanged.
		m := qInvNeg * t0
		u0, lo0 = bits.Mul64(m, q0)
		u1, lo1 = bits.Mul64(m, q1)
		u2, lo2 = bits.Mul64(m, q2)
		u3, lo3 = bits.Mul64(m, q3)
		u4, lo4 = bits.Mul64(m, q4)
		u5, lo5 = bits.Mul64(m, q5)
		_, c0 = bits.Add64(t0, lo0, 0)
		t0, c0 = bits.Add64(t1, lo1, c0)
		t1, c0 = bits.Add64(t2, lo2, c0)
		t2, c0 = bits.Add64(t3, lo3, c0)
		t3, c0 = bits.Add64(t4, lo4, c0)
		t4, c0 = bits.Add64(0, lo5, c0)
		u5, _ = bits.Add64(u5, 0, c0)
		t0, c0 = bits.Add64(u0, t0, 0)
		t1, c0 = bits.Add64(u1, t1, c0)
		t2, c0 = bits.Add64(u2, t2, c0)
		t3, c0 = bits.Add64(u3, t3, c0)
		t4, c0 = bits.Add64(u4, t4, c0)
		c2, _ = bits.Add64(c2, 0, c0)
		t4, c0 = bits.Add64(t5, t4, 0)
		t5, _ = bits.Add64(u5, c2, c0)
	}
	z[0] = t0
	z[1] = t1
	z[2] = t2
	z[3] = t3
	z[4] = t4
	z[5] = t5

	// if z ⩾ q → z -= q
	if !z.smallerThanModulus() {
		var b uint64
		z[0], b = bits.Sub64(z[0], q0, 0)
		z[1], b = bits.Sub64(z[1], q1, b)
		z[2], b = bits.Sub64(z[2], q2, b)
		z[3], b = bits.Sub64(z[3], q3, b)
		z[4], b = bits.Sub64(z[4], q4, b)
		z[5], _ = bits.Sub64(z[5], q5, b)
	}
	return z
}

// MulBy3 x *= 3 (mod q)
func MulBy3(x *Element) {
	_x := *x
	x.Double(x).Add(x, &_x)
}

// MulBy5 x *= 5 (mod q)
func MulBy5(x *Element) {
	_x := *x
	x.Double(x).Double(x).Add(x, &_x)
}

// MulBy13 x *= 13 (mod q)
func MulBy13(x *Element) {
	var y = Element{
		1176283927673829444,
		14130787773971430395,
		11354866436980285261,
		15740727779991009548,
		14951814113394531041,
		33013799364667434,
	}
	x.Mul(x, &y)
}

// Butterfly sets
//
//	a = a + b (mod q)
//	b = a - b (mod q)
//
//go:noescape
func Butterfly(a, b *Element)

func fromMont(z *Element) {
	_fromMontGeneric(z)
}

func reduce(z *Element) {
	_reduceGeneric(z)
}
